/*
 * NBCT
 * Copyright (c) 2007-2008, Yoshitaka Kameya
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 * 
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

/*
 * The memory managers specific to matrices/vectors, which is based on
 * the one described in the K&R book (Chap. 8.7).
 *
 * [Note]
 * When running with OpenMP (i.e. _OPENMP is defined), a matrix/vector
 * that is created by a thread (with alloc_matrix/vector()) should be
 * released by the same thread (with release_matrix/vector()).
 */

#ifdef _OPENMP
static M_HEADER **matrix_pool_;
static int *max_mpool_size_;
static M_HEADER **mp_free_;
static M_HEADER **mp_last_;

static V_HEADER **vector_pool_;
static int *max_vpool_size_;
static V_HEADER **vp_free_;
static V_HEADER **vp_last_;
#else
static M_HEADER *matrix_pool_;
static int max_mpool_size_;
static M_HEADER *mp_free_;
static M_HEADER *mp_last_;

static V_HEADER *vector_pool_;
static int max_vpool_size_;
static V_HEADER *vp_free_;
static V_HEADER *vp_last_;
#endif

#ifdef _OPENMP
/* `th' should be declared as a local variable */
#  define matrix_pool    (matrix_pool_[th])
#  define max_mpool_size (max_mpool_size_[th])
#  define mp_free        (mp_free_[th])
#  define mp_last        (mp_last_[th])
#  define vector_pool    (vector_pool_[th])
#  define max_vpool_size (max_vpool_size_[th])
#  define vp_free        (vp_free_[th])
#  define vp_last        (vp_last_[th])
#else
#  define matrix_pool    matrix_pool_
#  define max_mpool_size max_mpool_size_
#  define mp_free        mp_free_
#  define mp_last        mp_last_
#  define vector_pool    vector_pool_
#  define max_vpool_size max_vpool_size_
#  define vp_free        vp_free_
#  define vp_last        vp_last_
#endif

#ifdef _OPENMP
void initialize_matrix_pool(void)
{
  int nth = omp_get_max_threads();
  int th;

  matrix_pool_ = (M_HEADER **)n_malloc(sizeof(M_HEADER *)*nth);
  max_mpool_size_ = (int *)n_malloc(sizeof(int)*nth);
  mp_free_ = (M_HEADER **)n_malloc(sizeof(M_HEADER *)*nth);
  mp_last_ = (M_HEADER **)n_malloc(sizeof(M_HEADER *)*nth);

  for (th=0; th<nth; th++) {
    matrix_pool_[th] = (M_HEADER *)n_malloc(sizeof(M_HEADER)*M_POOL_SIZE);

    memset(matrix_pool_[th],0,sizeof(M_HEADER)*M_POOL_SIZE);

    max_mpool_size_[th] = M_POOL_SIZE;
    mp_free_[th] = NULL;
    mp_last_[th] = matrix_pool_[th];
  }
}
#else
void initialize_matrix_pool(void)
{
  matrix_pool_ = (M_HEADER *)n_malloc(sizeof(M_HEADER)*M_POOL_SIZE);

  memset(matrix_pool_,0,sizeof(M_HEADER)*M_POOL_SIZE);

  max_mpool_size_ = M_POOL_SIZE;
  mp_free_ = NULL;
  mp_last_ = matrix_pool_;
}
#endif

#ifdef _OPENMP
void initialize_vector_pool(void)
{
  int nth = omp_get_max_threads();
  int th;

  vector_pool_ = (V_HEADER **)n_malloc(sizeof(V_HEADER *)*nth);
  max_vpool_size_ = (int *)n_malloc(sizeof(int)*nth);
  vp_free_ = (V_HEADER **)n_malloc(sizeof(V_HEADER *)*nth);
  vp_last_ = (V_HEADER **)n_malloc(sizeof(V_HEADER *)*nth);

  for (th=0; th<nth; th++) {
    vector_pool_[th] = (V_HEADER *)n_malloc(sizeof(V_HEADER)*V_POOL_SIZE);

    memset(vector_pool_[th],0,sizeof(V_HEADER)*V_POOL_SIZE);

    max_vpool_size_[th] = V_POOL_SIZE;
    vp_free_[th] = NULL;
    vp_last_[th] = vector_pool_[th];
  }
}
#else
void initialize_vector_pool(void)
{
  vector_pool_ = (V_HEADER *)n_malloc(sizeof(V_HEADER)*V_POOL_SIZE);

  memset(vector_pool_,0,sizeof(V_HEADER)*V_POOL_SIZE);

  max_vpool_size_ = V_POOL_SIZE;
  vp_free_ = NULL;
  vp_last_ = vector_pool_;
}
#endif

/*
static void expand_matrix_pool(void)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  int new_max_mpool_size = max_mpool_size * 2;

  matrix_pool = (M_HEADER *)n_realloc(matrix_pool,
					sizeof(M_HEADER)*new_max_mpool_size);

  memset(mp_last,0,sizeof(M_HEADER)*(new_max_mpool_size - max_mpool_size));

  max_mpool_size = new_max_mpool_size;
}

static void expand_vector_pool(void)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  int new_max_vpool_size = max_vpool_size * 2;

  vector_pool = (V_HEADER *)n_realloc(vector_pool,
					sizeof(V_HEADER)*new_max_vpool_size);

  memset(vp_last,0,sizeof(V_HEADER)*(new_max_vpool_size - max_vpool_size));

  max_vpool_size = new_max_vpool_size;
}
*/

void release_matrix_pool(void)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  int n0,i,nunits;
  M_HEADER *mp;
  FLOAT_T **p;
  
  mp = matrix_pool;
  while (mp < mp_last) {
    n0 = mp->info.dim0;
    p = (FLOAT_T **)(mp + 1);
    for (i=0; i<n0; i++)
      free(p[i]);
    nunits = (n0 * sizeof(FLOAT_T *) + sizeof(M_HEADER) - 1)/sizeof(M_HEADER) + 1;
    mp += nunits;
  }

  free(matrix_pool);
  matrix_pool = NULL;
  mp_free = NULL;
  mp_last = NULL;
}

void release_vector_pool(void)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  free(vector_pool);
  vector_pool = NULL;
  vp_free = NULL;
  vp_last = NULL;
}

MATRIX alloc_matrix(int n0, int n1)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  int i,nunits;
  M_HEADER *mp,*mp_prev;
  FLOAT_T **p;

#ifdef CHECK_DIM
  if (n0 <= 0 || n1 <= 0)
    die("Internal error: alloc_matrix() -- illegal dimension (%d,%d)\n",n0,n1);
#endif
  
  nunits = (n0 * sizeof(FLOAT_T *) + sizeof(M_HEADER) - 1)/sizeof(M_HEADER) + 1;

  if ((mp_prev = mp_free) == NULL) {
    matrix_pool->info.next = mp_free = mp_prev = matrix_pool;
    matrix_pool->info.dim0 = 0;
    matrix_pool->info.dim1 = 0;
    mp_last = matrix_pool + 1;
  }

  mp = mp_prev->info.next;
  while (1) {
    if (mp->info.dim0 == n0 && mp->info.dim1 == n1) {
      mp_prev->info.next = mp->info.next;
      mp_free = mp_prev;
      return (MATRIX)(mp + 1);
    }

    if (mp == mp_free) {
      if ((mp_last + nunits - matrix_pool) >= max_mpool_size)
	die("Ran out of memory for matrices (allocated size: %d bytes)\n",max_mpool_size);
	/* expand_matrix_pool(); */

      mp_last->info.dim0 = n0;
      mp_last->info.dim1 = n1;

      p = (FLOAT_T **)(mp_last + 1);
      for (i=0; i<n0; i++) {
	p[i] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*n1);
      }
      
      mp_last += nunits;

      return (MATRIX)p;
    }

    mp_prev = mp;
    mp = mp->info.next;
  }
}

VECTOR alloc_vector(int n)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  V_HEADER *vp,*vp_prev;
  int nunits;
  VECTOR p;

#ifdef CHECK_DIM
  if (n <= 0)
    die("Internal error: alloc_vector() -- illegal dimension (%d)\n",n);
#endif
  
  nunits = (n * sizeof(FLOAT_T) + sizeof(V_HEADER) - 1)/sizeof(V_HEADER) + 1;

  if ((vp_prev = vp_free) == NULL) {
    vector_pool->info.next = vp_free = vp_prev = vector_pool;
    vector_pool->info.dim = 0;
    vp_last = vector_pool + 1;
  }

  vp = vp_prev->info.next;
  while (1) {
    if (vp->info.dim == n) {
      vp_prev->info.next = vp->info.next;
      vp_free = vp_prev;

      return (VECTOR)(vp + 1);
    }

    if (vp == vp_free) {
      if ((vp_last + nunits - vector_pool) >= max_vpool_size)
	die("Ran out of memory for vectors (allocated size: %d bytes)\n",max_vpool_size);
	/* expand_vector_pool(); */

      vp_last->info.dim = n;
      p = (VECTOR)(vp_last + 1);
      vp_last += nunits;

      return p;
    }

    vp_prev = vp;
    vp = vp->info.next;
  }
}

void set_matrix(MATRIX m, FLOAT_T **m0)
{
  int i,n0=-1,n1=-1;

#ifdef CHECK_DIM
  if ((n0 = get_dim0(m)) <= 0 || (n1 = get_dim1(m)) <= 0)
    die("Internal error: set_matrix() -- illegal dimension(s)\n");
#else
  n0 = get_dim0(m);
  n1 = get_dim1(m);
#endif

  for (i=0; i<n0; i++)
    memcpy(m[i],m0[i],sizeof(FLOAT_T)*n1);
}

void set_vector(VECTOR v, FLOAT_T *v0)
{
  int n;

#ifdef CHECK_DIM
  if ((n = get_dim(v)) <= 0)
    die("Internal error: set_vector() -- illegal dimension\n");
#else
  n = get_dim(v);
#endif

  memcpy(v,v0,sizeof(FLOAT_T)*n);
}

MATRIX alloc_set_matrix(int n0, int n1, FLOAT_T **m0)
{
  MATRIX m;

#ifdef CHECK_DIM
  if (n0 <= 0 || n1 <= 0)
    die("Internal error: alloc_set_matrix() -- illegal dimension(s)\n");
#endif

  m = alloc_matrix(n0,n1);
  set_matrix(m,m0);

  return m;
}

VECTOR alloc_set_vector(int n, FLOAT_T *v0)
{
  VECTOR v;
  
#ifdef CHECK_DIM
  if (n <= 0)
    die("Internal error: alloc_set_vector() -- illegal dimension(s)\n");
#endif

  v = alloc_vector(n);
  set_vector(v,v0);

  return v;
}

void copy_matrix(MATRIX m_dst, MATRIX m_src)
{
  int i,n0=-1,n1=-1;

#ifdef CHECK_DIM
  if ((n0 = get_dim0(m_dst)) <= 0 || (n1 = get_dim1(m_dst)) <= 0 ||
      n0 != get_dim0(m_src) || n1 != get_dim1(m_src))
    die("Internal error: copy_matrix() -- illegal dimension(s)\n");
#else
  n0 = get_dim0(m_dst);
  n1 = get_dim1(m_dst);
#endif

  for (i=0; i<n0; i++)
    memcpy(m_dst[i],m_src[i],sizeof(FLOAT_T)*n1);
}

void copy_vector(VECTOR v_dst, VECTOR v_src)
{
  int n;

#ifdef CHECK_DIM
  if ((n = get_dim(v_dst)) <= 0 || n != get_dim(v_src))
    die("Internal error: copy_vector() -- illegal dimension(s)\n");
#else
  n = get_dim(v_dst);
#endif

  memcpy(v_dst,v_src,sizeof(FLOAT_T)*n);
}

VECTOR dup_vector(VECTOR v)
{
  int n;
  VECTOR new_v;

#ifdef CHECK_DIM
  if ((n = get_dim(v)) <= 0)
    die("Internal error: dup_vector() -- illegal dimension(s)\n");
#else
  n = get_dim(v);
#endif
  
  new_v = alloc_vector(get_dim(v));
  copy_vector(new_v,v);

  return new_v;
}

MATRIX dup_matrix(MATRIX m)
{
  int n0=-1,n1=-1;
  MATRIX new_m;

#ifdef CHECK_DIM
  if ((n0 = get_dim0(m)) <= 0 || (n1 = get_dim1(m)) <= 0)
    die("Internal error: dup_matrix() -- illegal dimension(s)\n");
#else
  n0 = get_dim0(m);
  n1 = get_dim1(m);
#endif

  new_m = alloc_matrix(n0,n1);
  copy_matrix(new_m,m);
  
  return new_m;
}

void set_vprod_matrix(MATRIX m, VECTOR v0, VECTOR v1)
{
  int n0=-1,n1=-1,i,j;

#ifdef CHECK_DIM
  if ((n0 = get_dim(v0)) <= 0 || (n1 = get_dim(v1)) <= 0 ||
      n0 != get_dim0(m) || n1 != get_dim1(m))
    die("Internal error: set_vprod_matrix() -- illegal dimension(s)\n");
#else
  n0 = get_dim(v0);
  n1 = get_dim(v1);
#endif

  for (i=0; i<n0; i++)
    for (j=0; j<n1; j++)
      m[i][j] = v0[i] * v1[j];
}

MATRIX alloc_vprod_matrix(VECTOR v0, VECTOR v1)
{
  int n0=-1,n1=-1;
  MATRIX m;

#ifdef CHECK_DIM
  if ((n0 = get_dim(v0)) <= 0 || (n1 = get_dim(v1)) <= 0)
    die("Internal error: alloc_vprod_matrix() -- illegal dimension(s)\n");
#else
  n0 = get_dim(v0);
  n1 = get_dim(v1);
#endif
  
  m = alloc_matrix(n0,n1);
  set_vprod_matrix(m,v0,v1);

  return m;
}

VECTOR alloc_proj_vector(VECTOR v, int *use)
{
  return alloc_proj_vector_n(v,-1,use);
}

VECTOR alloc_proj_vector_n(VECTOR v, int n_use, int *use)
{
  int i,n=-1;
  VECTOR v_new;

#ifdef CHECK_DIM
  if ((n = get_dim(v)) <= 0 || n_use > n)
    die("Internal error: alloc_proj_vector_n() -- illegal dimension(s)\n");
#else
  n = get_dim(v);
#endif

  if (n_use < 0) {
    n_use = 0;
    for (i=0; i<n; i++) if (use[i]) n_use++;
  }

  v_new = alloc_vector(n_use);
  proj_vector(v,use,v_new);
  
  return v_new;
}

MATRIX alloc_proj_matrix(MATRIX m, int *use0, int *use1)
{
  return alloc_proj_matrix_n(m,-1,-1,use0,use1);
}

MATRIX alloc_proj_matrix_n(MATRIX m, int n_use0, int n_use1, int *use0, int *use1)
{
  int i=0,j=0,n0=-1,n1=-1;
  MATRIX m_new;

#ifdef CHECK_DIM
  if ((n0 = get_dim0(m)) <= 0 || (n1 = get_dim(m)) <= 0 || n_use0 > n0 || n_use1 > n1)
    die("Internal error: alloc_proj_matrix_n() -- illegal dimension(s)\n");
#else
  n0 = get_dim0(m);
  n1 = get_dim1(m);
#endif

  if (n_use0 < 0) {
    n_use0 = 0;
    for (i=0; i<n0; i++) if (use0[i]) n_use0++;
  }
  if (n_use1 < 0) {
    n_use1 = 0;
    for (j=0; j<n1; j++) if (use1[i]) n_use1++;
  }

  m_new = alloc_matrix(n_use0,n_use1);
  proj_matrix(m,use0,use1,m_new);
  
  return m_new;
}

void clear_matrix(MATRIX m)
{
  int n0=-1,n1=-1,i;

#ifdef CHECK_DIM
  if ((n0 = get_dim0(m)) <= 0 || (n1 = get_dim1(m)) <= 0)
    die("Internal error: clear_matrix() -- illegal dimension(s)\n");
#else
  n0 = get_dim0(m);
  n1 = get_dim1(m);
#endif

  for (i=0; i<n0; i++)
    memset(m[i],0,sizeof(FLOAT_T)*n1);
}

void clear_vector(VECTOR v)
{
  int n;

#ifdef CHECK_DIM
  if ((n = get_dim(v)) <= 0)
    die("Internal error: clear_vector() -- illegal dimension(s)\n");
#else
  n = get_dim(v);
#endif

  memset(v,0,sizeof(FLOAT_T)*n);
}

VECTOR alloc_clear_vector(int n)
{
  VECTOR v;

#ifdef CHECK_DIM
  if (n <= 0)
    die("Internal error: alloc_clear_vector() -- illegal dimension(s)\n");
#endif

  v = alloc_vector(n);
  clear_vector(v);

  return v;
}

MATRIX alloc_clear_matrix(int n0, int n1)
{
  MATRIX m;

#ifdef CHECK_DIM
  if (n0 <= 0 || n1 <= 0)
    die("Internal error: alloc_clear_matrix() -- illegal dimension(s)\n");
#endif

  m = alloc_matrix(n0,n1);
  clear_matrix(m);

  return m;
}

void release_matrix(MATRIX m)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  M_HEADER *mp,*b;

  if (m == NULL) return;

  b = (M_HEADER *)m - 1;

#ifdef CHECK_DIM
  if (b < matrix_pool || b > mp_last)
    die("Internal error: Illegal address of a matrix to be freed\n");
#endif

  mp = mp_free;
  while (1) {
    if (b == mp) die("Internal error: double-freed a matrix (%p)\n",b);
    if ((b > mp && b < mp->info.next) ||
	(mp >= mp->info.next && (b > mp || b < mp->info.next))) break;
    mp = mp->info.next;
  }

  b->info.next = mp->info.next;
  mp->info.next = b;

  mp_free = b;
}

void release_vector(VECTOR v)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  V_HEADER *vp,*b;

  if (v == NULL) return;

  b = (V_HEADER *)v - 1;

#ifdef CHECK_DIM
  if (b < vector_pool || b > vp_last)
    die("Internal error: Illegal address of a vector to be freed\n");
#endif

  vp = vp_free;
  while (1) {
    if (b == vp) die("Internal error: double-freed a vector (%p)\n",b);
    if ((b > vp && b < vp->info.next) ||
	(vp >= vp->info.next && (b > vp || b < vp->info.next))) break;
    vp = vp->info.next;
  }

  b->info.next = vp->info.next;
  vp->info.next = b;

  vp_free = b;
}

#ifdef DISABLE_DIM

int f_get_dim0(MATRIX m)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  M_HEADER *b;
  
  if (m == NULL) return 0;

  b = (M_HEADER *)m - 1;

#ifdef CHECK_DIM
  if (b < matrix_pool || b > mp_last) {
    fprintf(stderr,"Internal error: get_dim0()"
	    " -- illegal address (%p) of a matrix (pool = %p, last = %p)\n",
	    b,vector_pool,vp_last);
    exit(1);
  }
#endif

  return b->info.dim0;
}

int f_get_dim1(MATRIX m)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  M_HEADER *b;
  
  if (m == NULL) return 0;

  b = (M_HEADER *)m - 1;

#ifdef CHECK_DIM
  if (b < matrix_pool || b > mp_last) {
    fprintf(stderr,"Internal error: get_dim1()"
	    " -- illegal address (%p) of a matrix (pool = %p, last = %p)\n",
	    b,vector_pool,vp_last);
    exit(1);
  }
#endif

  return b->info.dim1;
}

int f_get_dim(VECTOR v)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  V_HEADER *b;

  if (v == NULL) return 0;

  b = (V_HEADER *)v - 1;

#ifdef CHECK_DIM
  if (b < vector_pool || b > vp_last) {
    fprintf(stderr,"Internal error: get_dim()"
	    " -- illegal address (%p) of a vector (vector_pool = %p, vp_last = %p)\n",
	    b,vector_pool,vp_last);
    exit(1);
  }
#endif

  return b->info.dim;
}

#endif /* DISABLE_DIM */
