/*
 * NBCT
 * Copyright (c) 2007-2008, Yoshitaka Kameya
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 * 
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

#define DESC_ENTRY_SIZE 128
#define DESC_COL_SIZE 3
#define MAX_ENTRY_ORDER 10

static char **desc_col;
static int desc_col_size;
static int max_desc_col_size;

static void initialize_desc_col(void)
{
  int i;

  desc_col = (char **)n_malloc(sizeof(char *)*DESC_COL_SIZE);
  for (i=0; i<DESC_COL_SIZE; i++)
    desc_col[i] = alloc_str(DESC_ENTRY_SIZE);

  desc_col_size = 0;
  max_desc_col_size = DESC_COL_SIZE;
}

static void expand_desc_col(void)
{
  int i,old_size,new_size;

  old_size = max_desc_col_size;
  new_size = old_size * 2;

  desc_col = (char **)n_realloc(desc_col,sizeof(char *)*new_size);
  for (i=old_size; i<new_size; i++)
    desc_col[i] = alloc_str(DESC_ENTRY_SIZE);

  max_desc_col_size = new_size;
}

static void release_desc_col(void)
{
  int i;
  for (i=0; i<max_desc_col_size; i++)
    free(desc_col[i]);
  free(desc_col);
  desc_col = NULL;
}

void initialize_desc(void)
{
  initialize_desc_col();
}

void release_desc(void)
{
  release_desc_col();
}

void print_usage(FILE *fp)
{
  fprintf(fp,conv_msg(usage));
}

void print_usage_full(FILE *fp)
{
  fprintf(fp,conv_msg(usage_full));
}

void print_id(FILE *fp, char *fmt, ...)
{
  va_list ap;

  fprintf(fp,"#### ID:");
  va_start(ap,fmt);
  vfprintf(fp,conv_desc(fmt),ap); fflush(fp_log);
  va_end(ap);

#ifdef USE_CRLF
  fprintf(fp,"\r\n");
#else
  fprintf(fp,"\n");
#endif
}

void print_desc(FILE *fp)
{
  int i;

  fprintf(fp,"#### ");

  for (i=0; i<desc_col_size; i++) {
    if (i > 0) fprintf(fp,", ");
    fprintf(fp,"%s",desc_col[i]);
  }

  desc_col_size = 0;

#ifdef USE_CRLF
  fprintf(fp,"\r\n####\r\n");
#else
  fprintf(fp,"\n####\n");
#endif
}

void print_desc_plain(FILE *fp, char *fmt, ...)
{
  va_list ap;

  fprintf(fp,"#### ");
  va_start(ap,fmt);
  vfprintf(fp,conv_desc(fmt),ap); fflush(fp_log);
  va_end(ap);
#ifdef USE_CRLF
  fprintf(fp,"\r\n####\r\n");
#else
  fprintf(fp,"\n####\n");
#endif
}

void print_data(FILE *fp, char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  vfprintf(fp,conv_data(fmt),ap);
  va_end(ap);

#ifdef USE_CRLF
  fprintf(fp,"\r\n");
#else
  fprintf(fp,"\n");
#endif
}

void print_data_aux(FILE *fp, int close, char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  vfprintf(fp,conv_data(fmt),ap);
  va_end(ap);

#ifdef USE_CRLF
  if (close) fprintf(fp,"\r\n");
#else
  if (close) fprintf(fp,"\n");
#endif
}

void add_desc(char *fmt, ...)
{
  va_list ap;
  char *s;

  if (desc_col_size == max_desc_col_size) expand_desc_col();
  s = desc_col[desc_col_size];

  va_start(ap,fmt);
#ifdef _MSC_VER
  _vsnprintf(s,DESC_ENTRY_SIZE,conv_desc(fmt),ap);
  s[DESC_ENTRY_SIZE - 1] = '\0';
#else
  vsnprintf(s,DESC_ENTRY_SIZE,conv_desc(fmt),ap);
#endif
  va_end(ap);

  desc_col_size++;
}

/* [Note]
 * This function returns a string allocated in the static area.
 * So we should make a copy of the returned string if necessary.
 */
char *get_case_id(int e)
{
  static char s[INT2STR_BUFSIZE];

  if (mode_id)
    return case_id_list[freq_list[e].case_id];
  else {
    strcpy(s,int2str(e));
    return s;
  }
}

char *freq_entry_val(int e, int v)
{
  static char *c = NULL;
  char *p = NULL;
  int u,a,v_d;

  if (attr_info[v].is_cont) {
    if (c == NULL) c = alloc_str(MAX_FLOAT_LEN);

    u = attr_info[v].gauss_id;
    a = attr_info[v].local_id;

    if (freq_list[e].miss_g[u][a])
      conv_sprintf(c,"?");
    else
      conv_snprintf(c,MAX_FLOAT_LEN,"%.Ng",freq_list[e].gval[u][a]);

    p = c;
  }
  else {
    v_d = attr_info[v].local_id;

    if (freq_list[e].id[v_d] >= 0)
      p = word_list[v_d][freq_list[e].id[v_d]];
    else if (freq_list[e].miss_w[v_d])
      p = "?";
    else die("Internal error: freq_entry_val(): word not assigned");
  }

  return p;
}
